#ifndef CLASS_TREE_H
#define CLASS_TREE_H

#include "incs.h"
#include <stddef.h>
#include "node.h"
#include "network.h"
#include "tree.h"

class Class_Tree : public Tree
{
public:
  Class_Tree(int n_batch, FILE* fp, int debug);
  Class_Tree(const Class_Tree& old);
  Class_Tree(Class_Tree* old);
  Class_Tree(int n_batch, 
		   int num_expert, 
		   int num_depth,
		   int num_ip, 
		   int num_op,
		   int debug,
		   int xprt_optim,
		   int gate_optim,
		   real _tol,
		   int _itmax,
		   int _msteps,
		   int _rich,
		   int _seed,
		   int _gate_ensemble = 0,
		   int _xprt_ensemble = 0,
		   int _gate_hyper_epochs = 10,
		   int _xprt_hyper_epochs = 10
		   );

  ~Class_Tree();
  Node* build_1layer_tree(Node* parent,
						  int num_train, 
						  int xprt_optim,
						  int gate_optim);

  Node* build_recursive_tree(Node* parent,
				 int num_experts, 
				 int* current_depth, 
				 int num_train, 
				 int xprt_optim,
				 int gate_optim);

  int update(int n_epochs, 
			 int init_epoch,
			 real** inputs, 
			 real** outputs, 
			 real** targets,
			 FILE* fp_res, 
			 int ntrain);
  int cv_update(int n_epochs, 
				int init_epoch,
				real** trn_inputs, 
				real** trn_targets,
				real** tst_inputs, 
				real** tst_targets,
				FILE* fp_res, 
				int ntrain,
				int ntest);
  int ens_update(int n_epochs, 
				 int init_epoch,
				 real** inputs, 
				 real** outputs, 
				 real** targets,
				 FILE* fp_res, 
				 int ntrain);
  void write_outputs(real** inputs, real** outputs, 
					 real** targets,
					 real** mom2, FILE* fp_res, int ntrain);
  void print_stats(FILE* fp, int ntrain, real** outputs, real** targets);
 double compute_like(int N, real** out, real** targ);
double compute_error(int N, real** out, real** targ);

private:
  int gate_ensemble;
  int xprt_ensemble;
  int gate_hyper_epochs;
  int xprt_hyper_epochs;
  real xprt_rand_offset;
  real gate_rand_offset;
};

#endif
