#include "utils.h"
#include "functions.h"
#include "incs.h"

#include <math.h>

#define MAX_SAFE_LIMIT 32.0


#ifndef isinf

int isinf(real arg){
  if(arg > FLT_MAX)
	return 1;
  else
	return 0;
}
#endif

static real minLogExp = -log(-LZERO);

real** remap_matrix_rows(real** in_vec, int dim, int dim2,
						 int* map) 
{
  real** out_vec = RealMatrix2(dim, dim2);
  for(int i = 0; i < dim; i ++) {
	int out_i = map[i];
	for(int d = 0; d < dim2; d ++) {
	  out_vec[i][d] = in_vec[out_i][d];
	}
  }
  return out_vec;
}

void remap_matrix_rows2(real** in_vec, int dim, int dim2,
						int* map, real** out_vec) 
{
  for(int i = 0; i < dim; i ++) {
	int out_i = map[i];
	for(int d = 0; d < dim2; d ++) {
	  out_vec[i][d] = in_vec[out_i][d];
	}
  }
}

	
real* randomize_vector(real* in_vec, int dim, int* map) {
  real* out_vec = new real[dim];
  int* obs = new int[dim];
  for(int d = 0; d < dim; d ++)
	obs[d] = 0;
  int out_i = 0;
  while(out_i < dim) {
	int guess = (int) ((1.0*dim)*get_drand());

	if(guess < dim) {
	  if(obs[guess] == 0) {
		obs[guess] = 1;
		map[out_i] = guess;
		out_vec[out_i] = in_vec[guess];
		out_i ++;
	  }
	}
	else {
	  error(1, "guess too big %d", guess);
	}
  }
  return out_vec;
}

void randomize_matrix_rows2(real** in_vec, int dim, int dim2,
							int* map, real** out_vec) {
  int* obs = new int[dim];
  for(int d = 0; d < dim; d ++)
	obs[d] = 0;
  int out_i = 0;
  while(out_i < dim) {
	int guess = (int) (dim*get_drand());
	if(guess < dim) {
	  if(obs[guess] == 0) {
		obs[guess] = 1;
		map[out_i] = guess;
		for(int d2 = 0; d2 < dim2; d2 ++) {
		  out_vec[out_i][d2] = in_vec[guess][d2];
		}
		out_i ++;
	  }
	}
  }
}

real** randomize_matrix_rows(real** in_vec, int dim, int dim2,
							  int* map) {
  real** out_vec = RealMatrix2(dim, dim2);
  int* obs = new int[dim];
  for(int d = 0; d < dim; d ++)
	obs[d] = 0;
  int out_i = 0;
  while(out_i < dim) {
	int guess = (int) (dim*get_drand());
	if(guess < dim) {
	  if(obs[guess] == 0) {
		obs[guess] = 1;
		map[out_i] = guess;
		for(int d2 = 0; d2 < dim2; d2 ++) {
		  out_vec[out_i][d2] = in_vec[guess][d2];
		}
		out_i ++;
	  }
	}
  }
  return out_vec;
}

	  

real gaussian_full(real *input,real *mean, real **inv_cov,real det,int n) {
  int i,j;
  real prob;
  real arg = 0;
  real error1;
  real *icov;
  
  for(i=0;i<n;i++) 
  {
	  icov = inv_cov[i];
	  error1 = input[i] - mean[i];
	  for(j=0;j<n;j++) 
	  {
		  arg += error1 * icov[j] * (input[j] - mean[j]);
	  }
  }
/*  if(isinf(arg)) error("arg is inf in gaussian %f",arg);*/
/*  if(isnan(arg)) error("arg is NaN in gaussian");*/
  prob = -0.5 * (arg + det);  
  /* assumes det is a log(det) */
  return prob;
}

real gaussian_diag(real *input,real *mean, real *inv_cov,real det,int n) {
  int i;
  real prob;
  real arg = 0;
  real error1 = 0;

  for(i=0;i<n;i++) {
    error1 = input[i] - mean[i];
    arg +=  error1 * inv_cov[i] * error1;
  }
/*  if(isinf(arg)) error("arg is inf in gaussian");*/
  prob = -0.5 *(arg + det); /* assumes det is a log(det) */
  return prob;
}

real gaussian_var(real *input,real *mean,real var,int n) {
  int i;
  real prob;
  real arg = 0;
  real error1 = 0;
  
  for(i=0;i<n;i++) {
    error1 = input[i] - mean[i];
    arg += error1 * error1;
  }
/*  if(isinf(arg))
    error("arg is inf in gaussian");*/
  prob = -0.5 * (arg + ( log(var)));
  return prob;
}

/* need a safety on the argument to this guy */
real max_vector(real *vector,int n) {
  int i;
  real max = -FLT_MAX;
   
  for(i=0;i<n;i++) 
    if(vector[i] > max) max = vector[i];
  return max;
}
void scale_vector(real *vector,real scale,int n) {
  int i;
  real sc;
  
  sc = 1.0 / scale;
  
  for(i=0;i<n;i++) vector[i] *= sc;
}
real trace_vector(real *vector,int n) {
  int i;
  real trace = 0.0;
  for(i=0;i<n;i++) 
    trace += vector[i];
  return trace;
}

void softmax(real *output, real *arg, int n) {
  int c; 
  real sum = 0.0; 
  real max = -FLT_MAX;

  max = max_vector(arg,n);

  for(c=0; c<n; c++) arg[c] -= max; 

  for(c=0; c<n; c++) 
    {
      if(arg[c] > MAX_SAFE_LIMIT)
		arg[c] = MAX_SAFE_LIMIT;
      if(arg[c] < - MAX_SAFE_LIMIT)
		arg[c] = - MAX_SAFE_LIMIT;
      output[c]   = exp(arg[c]); 
    }
  sum = trace_vector(output,n);
  scale_vector(output,sum,n);
}


void limit_outputs(real *output, int opdim, real upper, real lower) {
  int i;
  for(i=0;i<opdim;i++) {
    if(output[i] > upper) output[i] = upper;
    if(output[i] < lower) output[i] = lower;
/*    if(isnan(output[i])) error("output is NAN");*/
/*    if(isinf(output[i])) error("output is INF");*/
  }
}

real sigmoid(real arg) {
  real output;
  output = 1.0 / (1.0 + exp(-arg));
  return output;
}
real inner_product(real *vector1, real *vector2,int n) {
  
  int i;
  real prod = 0.0;;
  
  for(i=0; i<n; i++) 
    prod += vector1[i] * vector2[i];

  return prod;
}  

void softmax_output(real **theta, real *net_output, real *input, int ipdim, int opdim) 
{
  int c; 
  
  for(c = 0;  c<opdim;  c++) 
	{
	  net_output[c]  =  inner_product(input, theta[c], ipdim);  
	}
  softmax(net_output, net_output, opdim);
}

void linear_output(real **theta,real *net_output,real *input, int ipdim, int opdim) 
{
     int c; 
     for(c = 0;c< opdim; c++) 
       net_output[c] =  inner_product(input, theta[c], ipdim);
}

real Spow(real arg1,real arg2) {
/*  if(isnan(arg1))
    error("Argument 1 to Spow() is NaN");
  if(isnan(arg2))
    error("Argument 2 to Spow() is NaN");*/
/*  if(isinf(arg1))
    error("Argument 1 to Spow() is Inf");
  if(isinf(arg2))
    error("Argument 2 to Spow() is Inf");*/
  return(pow(arg1,arg2));
}


void   check_arg(real arg, char *s)
{
     if(isinf(arg))
       error(1, "%s is inf",s);
     if(isnan(arg))
       error(1, "%s is NaN",s);
}


# define SMALL (1.0 / (1 << 10))
# define LIMIT (1.0 / (1 << 20))

# ifdef IERF_TEST
static real x = -0.994;
int main() {
#else
real ierf(real x) {
#endif

  real last = 1.0, curr = 0.0;

  while(fabs(last - curr) > LIMIT) {
    real erfap = erf(curr);
    real slope = (erf(curr + SMALL) - erf(curr)) / SMALL;

    last = curr;
    curr += (x - erfap) / slope;

# ifdef IERF_TEST
    printf("%f\t%f\n", curr, erf(curr));
# endif

  }

#ifdef IERF_TEST
  exit(0);
#else
  return(curr);
#endif

}


long get_seed()
{
  struct timeval t;
  struct timezone tz;
  tz.tz_dsttime = DST_NONE;
  tz.tz_minuteswest = 0;
  gettimeofday(&t,&tz);
  return t.tv_usec + t.tv_sec;
}
void set_seed(long seed)
{
  srand48(seed);
}

long get_random()
{
  return lrand48();
}

double get_drand()
{
  /*  couldn't get drand48() working on SUNs  */
  return (double)lrand48()/2147483648.0L ;   /*  (double)(~2**31);  */
}




/* EXPORT->LSub: Return diff x - y on log scale, 
                 diff < LSMALL is floored to LZERO */
real LSub(real x, real y)
{
   real diff,z;
   
   if (x<y)    
      error(1,"LSub: result -ve");
   diff = y-x;
   if (diff<minLogExp)
      return  (x<LSMALL)?LZERO:x;
   else {
      z = 1.0 - exp(diff);
      return (z<MINLARG) ? LZERO : x+log(z);
   }
}


real ran0(int* idum)
{
	static real y,maxran,v[98];
	real dum;
	static int iff=0;
	int j;
	unsigned i,k;

	if (*idum < 0 || iff == 0) {
		iff=1;
		i=2;
		do {
			k=i;
			i<<=1;
		} while (i);
		maxran=k;
		srand(*idum);
		*idum=1;
		for (j=1;j<=97;j++) dum=rand();
		for (j=1;j<=97;j++) v[j]=rand();
		y=rand();
	}
	j= (int)(1+97.0*y/maxran);
	if (j > 97 || j < 1) error(1,"RAN0: This cannot happen.");
	y=v[j];
	v[j]=rand();
	return y/maxran;
}

#define M1 259200
#define IA1 7141
#define IC1 54773
#define RM1 (1.0/M1)
#define M2 134456
#define IA2 8121
#define IC2 28411
#define RM2 (1.0/M2)
#define M3 243000
#define IA3 4561
#define IC3 51349

real ran1(int* idum)
{
	static long ix1,ix2,ix3;
	static real r[98];
	real temp;
	static int iff=0;
	int j;

	if (*idum < 0 || iff == 0) {
		iff=1;
		ix1=(IC1-(*idum)) % M1;
		ix1=(IA1*ix1+IC1) % M1;
		ix2=ix1 % M2;
		ix1=(IA1*ix1+IC1) % M1;
		ix3=ix1 % M3;
		for (j=1;j<=97;j++) {
			ix1=(IA1*ix1+IC1) % M1;
			ix2=(IA2*ix2+IC2) % M2;
			r[j]=(ix1+ix2*RM2)*RM1;
		}
		*idum=1;
	}
	ix1=(IA1*ix1+IC1) % M1;
	ix2=(IA2*ix2+IC2) % M2;
	ix3=(IA3*ix3+IC3) % M3;
	j= (int) (1 + ((97*ix3)/M3));
	if (j > 97 || j < 1) error(1,"RAN1: This cannot happen.");
	temp=r[j];
	r[j]=(ix1+ix2*RM2)*RM1;
	return temp;
}

#undef M1
#undef IA1
#undef IC1
#undef RM1
#undef M2
#undef IA2
#undef IC2
#undef RM2
#undef M3
#undef IA3
#undef IC3

#define M 714025
#define IA 1366
#define IC 150889

real ran2(int* idum)
{
	static long iy,ir[98];
	static int iff=0;
	int j;

	if (*idum < 0 || iff == 0) {
		iff=1;
		if ((*idum=(IC-(*idum)) % M) < 0) *idum = -(*idum);
		for (j=1;j<=97;j++) {
			*idum=(IA*(*idum)+IC) % M;
			ir[j]=(*idum);
		}
		*idum=(IA*(*idum)+IC) % M;
		iy=(*idum);
	}
	j = (int) (1 + 97.0*iy/M);
	if (j > 97 || j < 1) error(1,"RAN2: This cannot happen.");
	iy=ir[j];
	*idum=(IA*(*idum)+IC) % M;
	ir[j]=(*idum);
	return (real) iy/M;
}

#undef M
#undef IA
#undef IC


real gasdev(int *idum)
{
	static int iset=0;
	static real gset;
	real fac,r,v1,v2;

	if  (iset == 0) {
		do {
			v1=2.0*ran1(idum)-1.0;
			v2=2.0*ran1(idum)-1.0;
			r=v1*v1+v2*v2;
		} while (r >= 1.0);
		fac=sqrt(-2.0*log(r)/r);
		gset=v1*fac;
		iset=1;
		return v2*fac;
	} else {
		iset=0;
		return gset;
	}
}

real  Slog(real arg)
{
  if(arg < MINLARG)
    return LZERO;
  if(arg > MAXLARG)
    return LOGINF;
  return log(arg);
}

/* EXPORT->LAdd: Return sum x + y on log scale, 
                sum < LSMALL is floored to LZERO */

real    LAdd(real x, real y) {
  real temp,diff,z;
   
   if (x<y) {
      temp = x; x = y; y = temp;
   }
   diff = y-x;
   if (diff<minLogExp)
      return  (x<LSMALL)?LZERO:x;
   else {
      z = exp(diff);
      return x+log(1.0+z);
   }
}

void log_softmax(real *output, real *arg, int n_output) {
  double sum = LZERO;
  for(int i = 0; i < n_output; i ++) {
	if(arg[i] > LSMALL)
	  sum = LAdd(sum, arg[i]);
  }
  for(int j = 0; j < n_output; j ++) {
    arg[j] -= sum;
    output[j] = arg[j];
  }
}



#define M 7
#define NSTACK 50
#define FM 7875
#define FA 211
#define FC 1663

void qcksrt(int n, double* arr)
{
	int l=1,jstack=0,j,ir,iq,i;
	int istack[NSTACK+1];
	long int fx=0L;
	double a;

	ir=n;
	for (;;) {
		if (ir-l < M) {
			for (j=l+1;j<=ir;j++) {
				a=arr[j];
				for (i=j-1;arr[i]>a && i>0;i--) arr[i+1]=arr[i];
				arr[i+1]=a;
			}
			if (jstack == 0) return;
			ir=istack[jstack--];
			l=istack[jstack--];
		} else {
			i=l;
			j=ir;
			fx=(fx*FA+FC) % FM;
			iq=l+((ir-l+1)*fx)/FM;
			a=arr[iq];
			arr[iq]=arr[l];
			for (;;) {
				while (j > 0 && a < arr[j]) j--;
				if (j <= i) {
					arr[i]=a;
					break;
				}
				arr[i++]=arr[j];
				while (a > arr[i] && i <= n) i++;
				if (j <= i) {
					arr[(i=j)]=a;
					break;
				}
				arr[j--]=arr[i];
			}
			if (ir-i >= i-l) {
				istack[++jstack]=i+1;
				istack[++jstack]=ir;
				ir=i-1;
			} else {
				istack[++jstack]=l;
				istack[++jstack]=i-1;
				l=i+1;
			}
			if (jstack > NSTACK) error(1, "NSTACK too small in QCKSRT");
		}
	}
}

#undef M
#undef NSTACK
#undef FM
#undef FA
#undef FC

void indexx(int n, double* arrin,int* indx)
{
	int l,j,ir,indxt,i;
	float q;

	for (j=1;j<=n;j++) indx[j]=j;
	l=(n >> 1) + 1;
	ir=n;
	for (;;) {
		if (l > 1)
			q=arrin[(indxt=indx[--l])];
		else {
			q=arrin[(indxt=indx[ir])];
			indx[ir]=indx[1];
			if (--ir == 1) {
				indx[1]=indxt;
				return;
			}
		}
		i=l;
		j=l << 1;
		while (j <= ir) {
			if (j < ir && arrin[indx[j]] < arrin[indx[j+1]]) j++;
			if (q < arrin[indx[j]]) {
				indx[i]=indx[j];
				j += (i=j);
			}
			else j=ir+1;
		}
		indx[i]=indxt;
	}
}
#include <math.h>

float gammln(float xx)
{
        double x,tmp,ser;
        static double cof[6]={76.18009173,-86.50532033,24.01409822,
                -1.231739516,0.120858003e-2,-0.536382e-5};
        int j;

        x=xx-1.0;
        tmp=x+5.5;
        tmp -= (x+0.5)*log(tmp);
        ser=1.0;
        for (j=0;j<=5;j++) {
                x += 1.0;
                ser += cof[j]/x;
        }
        return -tmp+log(2.50662827465*ser);
}
