#include "incs.h"
#include "net.h"
#include "utils.h"
Net::~Net() {
  FreeRealMatrix2(&tmp_op_vec);
  FreeRealVector(&jpost);
  FreeRealVector(&tmp_op);
  FreeCharVector(&class_str);
}
Net::Net(int n_batch, 
	 int num_ip, 
	 int num_op,
	 int dbg)
  : size_batch(n_batch),
    n_input(num_ip),
    n_output(num_op),
    debug(dbg)
{
  tmp_op_vec = RealMatrix2(size_batch, n_output);
  jpost = RealVector(size_batch);
  // temp stuff
  tmp_op = RealVector(n_output);

  class_str = CharVector(1024);
}

Net::Net(const Net & old)
  : size_batch(old.size_batch),
    n_input(old.n_input),
    n_output(old.n_output),
    debug(old.debug),
	batch_size(old.batch_size)
{
  jpost = RealVector(size_batch);
  // temp stuff
  tmp_op = RealVector(n_output);
  tmp_op_vec = RealMatrix2(size_batch, n_output);
  likelihood = 0.0;
  class_str = CharVector(1024);
  strcpy(class_str, "Net");
}

Net::Net(int n_batch, FILE* fp) :
  size_batch(n_batch)
{
  class_str = CharVector(1024);
  fscanf(fp, "%d %d\n", &n_output, &n_input);
  jpost = RealVector(size_batch);
  // temp stuff
  tmp_op = RealVector(n_output);
  tmp_op_vec = RealMatrix2(size_batch, n_output);

} 

void Net::go_forward_batch(int N, real** in, real** out)
{
  for(int n = 0; n < N; n ++){
    go_forward(in[n], out[n]);
  }
}

void Net::go_forward_batch(int N, real** in, real** out, real** m2)
{
  for(int n = 0; n < N; n ++){
    go_forward(in[n], out[n], m2[n]);
  }
}
