#include "incs.h"
#include "net_class_mac.h"
#include "utils.h"
#include "functions.h"
#include "macopt.h"

NET_Class_Mac::NET_Class_Mac(int n_batch, 
		     int num_ip, 
		     int num_op,
		     int dbg,
		     int _itmax,
		     int _msteps,
		     int _rich,
		     real _tol)
  : itmax(_itmax),
    rich(_rich),
    NET_Class(n_batch, num_ip, num_op, dbg, _msteps, _tol),
    Macopt(num_ip*num_op, dbg, _rich)
{
  grad = RealVector(n_output * n_input);
  old_weights = RealVector(n_output * n_input);
  mac_weights = RealVector(n_output * n_input + 1);
  strcpy(class_str,"NET_Class_Mac");
}

NET_Class_Mac::NET_Class_Mac(const NET_Class_Mac & old)
  : itmax(old.itmax),
    rich(old.rich),
    NET_Class(old),
    Macopt(old.n_input * old.n_output, old.debug, old.rich)
{
  grad = RealVector(n_output * n_input);
  for(int i1 = 0; i1 < n_output*n_input; i1 ++) {
	grad[i1] = old.grad[i1];
  }
  old_weights = RealVector(n_output * n_input);
  for(int i2 = 0; i2 < n_output*n_input; i2 ++) {
	old_weights[i2] = old.old_weights[i2];
  }
  mac_weights = RealVector(n_output * n_input + 1);
  for(int i = 0; i < n_output*n_input+1; i ++) {
	mac_weights[i] = old.mac_weights[i];
  }
  strcpy(class_str,"NET_Class_Mac");
  //  printf("copying %s\n", class_str);
}

NET_Class_Mac::NET_Class_Mac(int n_batch, FILE* fp) 
  : NET_Class(n_batch, fp),
	Macopt(n_input*n_output, 0, 0)
{
  grad = RealVector(n_output * n_input);
  old_weights = RealVector(n_output * n_input);
  mac_weights = RealVector(n_output * n_input + 1);
}  
  
NET_Class_Mac::~NET_Class_Mac() 
{
  FreeRealVector(&grad);
  FreeRealVector(&mac_weights);
  FreeRealVector(&old_weights);
}

double NET_Class_Mac::train_params(int n_points)
{
  double cost = 0.0;
  batch_size = n_points;

  for(int n1 = 1; n1 <= a_n; n1 ++) {
    mac_weights[n1] = weights[n1-1];
  }
  // double epsilon=0.00001;
  //  printf("checking....\n");

  //  maccheckgrad( mac_weights, a_n, epsilon, 0 );

  macoptII(mac_weights, a_n, itmax, tol);

  for(int n = 1; n <= a_n; n ++) {
    weights[n-1] = mac_weights[n];
  }
  return cost;
}

double NET_Class_Mac::train_params(int n_points,  real* wts)
{
  batch_size = n_points;
  for(int p = 0; p < n_points; p ++)
	jpost[p] = wts[p];

  for(int n1 = 1; n1 <= a_n; n1 ++) {
    mac_weights[n1] = weights[n1-1];
  }
  double epsilon=0.00001;
  if(debug > 3) {
    printf("checking....\n");
    maccheckgrad( mac_weights, a_n, epsilon, 0 );
  }
  macoptII(mac_weights, a_n, itmax, tol);
  for(int n = 1; n <= a_n; n ++) {
    weights[n-1] = mac_weights[n];
  }
  double cost = 0.0;
  return cost;
}

double NET_Class_Mac::func(double* new_weights) {
  double fn = 0.0;
  for(int w = 0; w < n_weights; w ++) {
	old_weights[w] = weights[w];
  }

  for(int j = 1; j <= a_n; j ++) {
    weights[j-1] = new_weights[j];
  }
  fn = get_function();
  for(int w = 0; w < n_weights; w ++) {
	weights[w] = old_weights[w];
  }

  return fn;
}
double* NET_Class_Mac::get_grad(int n_points, real** in,
								real** targ) {
  batch_size = n_points;
  /*  for(int n2 = 0; n2 < n_points; n2 ++) {
	for(int i = 0; i < n_input; i ++) {
	  inputs[n2][i] = in[n2][i];
	}
	for(int o = 0; o < n_output; o ++) {
	  targets[n2][o] = targ[n2][o];
	}
  } */
  inputs = in;
  targets = targ;

  get_gradient(grad);
  return grad;
}

void NET_Class_Mac::dfunc(double* new_weights, double* dw) {
  for(int w = 0; w < n_weights; w ++) {
	old_weights[w] = weights[w];
  }

  for(int j1 = 1; j1 <= a_n; j1 ++) {
    weights[j1-1] = new_weights[j1];
  }
  get_gradient(grad);

  for(int j = 1; j <= a_n; j ++) {
    dw[j] = grad[j-1];
  }
  for(int w = 0; w < n_weights; w ++) {
	weights[w] = old_weights[w];
  }
}

