#include "incs.h"
#include "net_regress_ens_irls.h"
#include "chol.h"
#include "matrix.h"
#include "utils.h"
#include "functions.h"
#include "lu.h"

NET_Regress_Ens_Irls::NET_Regress_Ens_Irls(int n_batch, 
				   int num_ip, 
				   int num_op,
				   int dbg,
				   int _msteps,
				   real _tol,
				   real _lambda,
				   real _prior_var  )
  :
  NET_Regress(n_batch, num_ip, num_op, _lambda, _prior_var,
				dbg, _msteps, _tol)
{
  params = RealMatrix2(n_output, n_input);
  grad = RealMatrix2(n_output, n_input);
  grad2 = RealVector(n_output * n_input);
  hessian = RealMatrix3(n_output, n_input, n_input);
  learn_rate = 1.0;
  strcpy(class_str, "NET_Regress_Ens_Irls");
}

NET_Regress_Ens_Irls::NET_Regress_Ens_Irls(const NET_Regress_Ens_Irls& old)
  : NET_Regress(old)
{
  params = new real*[n_output];
  for(int i = 0; i < n_output; i ++){
    params[i] = new real[n_input];
  }
  for(int m = 0; m < n_output; m ++) {
	for(int i2 = 0; i2 < n_output; i2 ++){
	  params[m][i2] = old.params[m][i2];
	}
  }
  grad = RealMatrix2(n_output, n_input);
  grad2 = RealVector(n_output * n_input);
  for(int o = 0; o < n_output; o ++) {
	for(int j = 0; j < n_output; j ++){
	  grad[o][j] = old.grad[o][j];
	}
  }
  hessian = RealMatrix3(n_output, n_input, n_input);
  for(int k = 0; k < n_output; k ++) {
	for(int i3 = 0; i3 < n_output; i3 ++){
	  for(int i2 = 0; i2 < n_output; i2 ++){
		hessian[k][i3][i2] = old.hessian[k][i3][i2];
	  }
	}
  }
  learn_rate = 1.0;
  strcpy(class_str, "NET_Regress_Ens_Irls");
}

NET_Regress_Ens_Irls::~NET_Regress_Ens_Irls()
{
}

double* NET_Regress_Ens_Irls::get_grad(int n_points, real** in,
								  real** targ) {
  batch_size = n_points;
  inputs = in;
  targets = targ;
  get_gradient(grad2);
  return grad2;
}

double NET_Regress_Ens_Irls::train_params(int n_points)
{
  printf("using irls\n");
  double cost = 0.0;
  for(int op =0; op < n_output; op ++) {
    for(int ip =0; ip < n_input; ip ++) {
      grad[op][ip] = 0.0;
      for(int ip2 =0; ip2 < n_input; ip2 ++) {
		hessian[op][ip][ip2] = 0.0;
      }
    }
  }
  
  for(int point = 0; point < n_points; point ++){
    for(int op3 =0; op3 < n_output; op3 ++) {
      for(int ip =0; ip < n_input; ip ++) {
		grad[op3][ip] += targets[point][op3] * inputs[point][ip] * inv_variance[op3];
		for(int ip2 =0; ip2 < n_input; ip2 ++) {
		  hessian[op3][ip][ip2] +=  inputs[point][ip2] * inputs[point][ip] * inv_variance[op3];
		}
      }
    }
  }
  for(int op2 =0; op2 < n_output; op2 ++) {
    lin_solve(hessian[op2], grad[op2], n_input, params[op2]);
  }
  for(int op4 = 0; op4 < n_output; op4 ++) {
    for(int ip = 0; ip < n_input; ip ++) {
      weights[op4*n_input + ip] = params[op4][ip];
    }
  }
  return cost;
}

double NET_Regress_Ens_Irls::train_params(int n_points, 
									  real* wts)
{
  double cost = 0.0;
  for(int op =0; op < n_output; op ++) {
    for(int ip =0; ip < n_input; ip ++) {
      grad[op][ip] = 0.0;
      for(int ip2 =0; ip2 < n_input; ip2 ++) {
		hessian[op][ip][ip2] = 0.0;
      }
    }
  }
  
  for(int point = 0; point < n_points; point ++){
    for(int op5 =0; op5 < n_output; op5 ++) {
      real dink = wts[point] / variance[op5];
      for(int ip =0; ip < n_input; ip ++) {
	grad[op5][ip] +=  dink * targets[point][op5] * inputs[point][ip];	
 	for(int ip2 =0; ip2 < n_input; ip2 ++) {
	  hessian[op5][ip][ip2] +=  dink * inputs[point][ip2] * inputs[point][ip];
	}
      }
    }
  }
  for(int op2 =0; op2 < n_output; op2 ++) {
    lin_solve(hessian[op2], grad[op2], n_input, params[op2]);
    if(debug > 2){
      printf("params = \n");
      print_vector(params[op2], n_input);
    }
  }
  for(int op6 = 0; op6 < n_output; op6 ++) {
    for(int ip = 0; ip < n_input; ip ++) {
      weights[op6*n_input + ip] = params[op6][ip];
    }
  }
  return cost;
}
