#include "incs.h"
#include "node.h"
#include "utils.h"

Node::~Node() {
   FreeRealVector(&cond_post);
   FreeRealVector(&joint_post);
   FreeRealVector(&joint_prob);
   FreeRealVector(&cond_prob);
   FreeCharVector(&class_str);
}

Node::Node(const Node& old) 
  : parent(old.parent),
    n_input(old.n_input),
    n_output(old.n_output),
    id(old.id),
	size_batch(old.size_batch),
    debug(old.debug)
{
  cond_post  = RealVector(size_batch);
  joint_post = RealVector(size_batch);
  joint_prob = RealVector(size_batch); 
  cond_prob  = RealVector(size_batch);
  class_str = CharVector(1024);
  for(int point = 0; point < size_batch; point ++){
    cond_prob[point]  = old.cond_prob[point];
    joint_prob[point] = old.joint_prob[point];
    cond_post[point]  = old.cond_post[point];
	joint_post[point] = old.joint_post[point];
  }

  strcpy(class_str, "Node");
  visited = 0;
}
Node::Node(int n_batch, 
		   int num_ip, 
		   int num_op,
		   int dbg,
		   int node_id)
  : n_input(num_ip),
    n_output(num_op), 
    id(node_id),
	size_batch(n_batch),
    debug(dbg)
{
  cond_post  = RealVector(size_batch);
  joint_post = RealVector(size_batch);
  joint_prob = RealVector(size_batch); 
  cond_prob  = RealVector(size_batch);
  class_str = CharVector(1024);
  for(int point = 0; point < size_batch; point ++) {
    cond_post[point]  = 0.0;
	joint_post[point] = 0.0;
    cond_prob[point]  = 0.0;
	joint_prob[point] = 0.0;
  }

  visited = 0;
}

Node::Node(int n_batch, FILE* fp) 
  : size_batch(n_batch)
{
  class_str = CharVector(1024);
  cond_post  = RealVector(size_batch);
  joint_post = RealVector(size_batch);
  joint_prob = RealVector(size_batch);
  cond_prob  = RealVector(size_batch);
  for(int point = 0; point < size_batch; point ++) {
    cond_post[point]  = 0.0;
	joint_post[point] = 0.0;
    cond_prob[point]  = 0.0;
	joint_prob[point] = 0.0;
  }
  visited = 0;
}


