#include "tree.h"


void Tree::grow_train(int n_epochs, real** inputs, real** targets, FILE* fp_res, 
					  int npoints, int n_splits) {
  
  train(n_epochs, 0, inputs, targets, fp_res, npoints, 1);
  real last_err = FLT_MAX;
  real err = compute_error(npoints, inputs, targets);
  int iter = 0;
  while(((1.0 - err/last_err) > 0.001)||(iter < n_splits)) {
	printf("\n---------\n split %d\n -------\n", iter);
	split_nodes_hard(npoints, inputs, targets);
	train(n_epochs, 0, inputs, targets, fp_res, npoints, 1);	
	last_err = err;
	err = compute_error(npoints, inputs, targets);
	printf("\n iter %d last err = %f, err = %f, change = %f\n", iter, last_err, err,
		   1.0 - err/last_err);
	iter ++;
  }
} 

void Tree::grow_train(int n_epochs, 
					  int n_train, real** trn_inputs, real** trn_targets, 
					  int n_test, real** tst_inputs, real** tst_targets, 
					  FILE* fp_out) {
  
  //  train(n_epochs, 0, trn_inputs, trn_targets, fp_out, n_train, 1);
  write_weights(fp_out);
  real min_err = FLT_MAX;
  cv_train(n_epochs, 0, trn_inputs, trn_targets, 
		   tst_inputs, tst_targets, fp_out, n_train, n_test,
		   &min_err);	
  real last_err = FLT_MAX;
  real err = min_err;
  int iter = 0;
  while(((last_err > err) || (iter < 2)) && (iter < 50)) {
	printf("---------\n split %d\n -------\n", iter);
	//	split_nodes_fsearch(n_train, trn_inputs, trn_targets);
	split_nodes_hard(n_train, trn_inputs, trn_targets);
	//  split_nodes_max_like(n_train, trn_inputs, trn_targets);
	cv_train(n_epochs, 0, trn_inputs, trn_targets, 
			 tst_inputs, tst_targets, fp_out, n_train, n_test,
			 &min_err);	
	last_err = err;
	err = min_err;
	printf("\n iter %d last err = %f, err = %f, change = %f\n", iter, 
		   last_err, err,
		   1.0 - err/last_err);
	iter ++;
  }
  printf("----------------------end----------------\n");
} 
void Tree::split_nodes_fsearch(int n_points, real** in, real** targ) {
  Node_NT* best_node;
  real best_diff = -FLT_MAX;
  Node** trial_node = new Node*[n_terminals];
  int n_splits = 0;
  int best_split_num = 0;
  int best_node_num = 0;
  for(int term = 0; term < n_terminals; term ++) {
	Node* nd = term_nodes[term];
	printf("trying node %d %d\n", term, nd->get_id());
	real diff;
	int success;
	Node_NT* guess_node = find_best_soft_node(nd, n_points, in, targ,
											  &diff, &success);
	//	Node_NT* guess_node = find_best_hard_node(nd, n_points, in, targ,
	//											  &diff, &success);
	if(success == 1) {
	  if(diff >best_diff) {
		best_diff = diff;
		trial_node[n_splits] = guess_node;
		best_node = guess_node;
		best_split_num = n_splits;
		best_node_num = term;
		n_splits ++;
	  }	
	}
  }
  if(n_splits > 0) {
	printf("best_diff = %f %d\n", best_diff, best_node->get_id());
	
	if(n_gates == 0) {
	  delete node;
	  node = best_node;
	}
	else {
	  delete term_nodes[best_node_num];
	  swap_nodes(best_node, term_nodes[best_node_num]);
	}
	for(int t = 0; t < n_splits; t ++) {
	  if(t != best_split_num)
		delete_new_nt((Node_NT*)trial_node[t]);
	}
  }
  else {
	printf("no splits made\n");
  }
  delete [] trial_node;
  delete [] term_nodes;
  delete [] terminals;
  delete [] gates;
  delete [] nt_nodes;
  gather_nodes();
  mac_pass = 0;
  macopt_free();
  int np = get_n_params();
  printf("nparams = %d\n", np);

}

Node_NT* Tree::find_best_soft_node(Node* nd, int n_points, 
								   real** in, real** targ, 
								   real* glob_best_diff,
								   int* success) {
  real best_thresh = 0.0;
  int best_feat = -1;
  int n_divs = 4;
  Node** trial_node = new Node*[n_input*n_divs];
  Node_NT* best_node;
  int best_split_num = 0;
  int n_splits = 0;
  int batch_n = 0;
  real best_diff = -FLT_MAX;
  for(int n = 0; n < n_points; n ++) {
	real wt = Sexp(nd->get_joint_post(n));
	if(wt > 0.01) {
	  for(int i = 0; i < n_input; i ++) {
		in_batch[batch_n][i] = in[n][i];
	  }
	  batch_n ++;
	}
  }
  *success = 0;
  if(batch_n < 4) {
	printf("no split tried, batch_n < 4\n");
  }
  else {
	for(int i = 1; i < n_input; i ++) {
	  printf("searching dimension %d\n", i);
	  // for each feature dimension, sort observations
	  for(int n = 0; n < batch_n; n ++) {
		tmp[n+1] = in_batch[n][i];
	  }
	  indexx(batch_n, tmp, indx);
	  for(int n = 0; n < batch_n; n ++) {
		srt_in[n][i] = tmp[indx[n+1]];
	  }
	  // now try at most n_divs divisions of data
	  int num_divs = batch_n;
	  if(batch_n > n_divs) num_divs = n_divs;
	  int lim = (int) (batch_n / ((double) num_divs));
	  for(int d = 0; d < num_divs-1; d ++) {
		//		real thresh = (srt_in[lim*d][i] + srt_in[lim*(d+1)][i]) / 2.0;
		real thresh = srt_in[lim*(d+1)][i];
		
		//		assign_probs_from_thresh(nd, n_points, in, thresh, i, jpost, cpost);
		int thresh_posn = (int) (lim*(d+1));
		assign_probs_from_thresh_sig(nd, n_points, in, thresh, thresh_posn,
									 i, jpost, cpost, batch_n,
									 srt_in, indx);
		real old_like = get_node_aux(nd, n_points, in, targ);
		real sum_jp = get_node_sum_jpost(nd, n_points, in, targ);
		sum_jp = Sexp(sum_jp);
		Node_NT* new_nd = split_this_node(thresh, i, nd, n_points, jpost, cpost, 
										  in, targ);
		real new_like = get_node_aux(new_nd, n_points, in, targ);
		real diff = new_like - old_like;
		diff /= sum_jp;
		printf("sum_jp = %f new = %f old = %f diff = %f %d %f\n", 
			   sum_jp, new_like, old_like, diff, new_nd->get_id(), thresh);
		if(diff > best_diff) {
		  best_node = new_nd;
		  best_diff = diff;
		  best_thresh = thresh;
		  best_split_num = n_splits;
		  best_feat = i;
		  trial_node[n_splits] = new_nd;
		  n_splits ++;
		  *success = 1;
		}
		else {
		  //		  printf("deleting node\n");
		  delete_new_nt(new_nd);
		  new_nd = (Node_NT*) NULL;
		}
	  }
	}
  }
  *glob_best_diff = best_diff;
  for(int t = 0; t < n_splits; t ++) {
	if(t != best_split_num)
	  delete_new_nt((Node_NT*)trial_node[t]);
  }
  delete [] trial_node;
  return best_node;
}

Node* Tree::create_new_nt(Node* old_child, int num_children, Network* old_gate) {
  // if the child is a terminal node,
  // create new non terminal from this child
  Node_NT* new_child = new Node_NT(*old_child, num_children);

  // now sort out the gate type and create new gate
  // from gate of current node.

  Network* gate;
	
  gate = get_new_net(old_gate);

  // set the new gate to the gate of the new child

  new_child->set_gate(gate);
  
  // now set the terminal nodes of the new terminal
  // allocate some memory for them.
  
  Node** new_terminals = new Node*[num_children];
  for(int ch = 0; ch < num_children; ch ++) {
	
	Network* expert;
	Network* xprt = ((Node_Term*) old_child)->get_expert();

	expert = get_new_net(xprt);

	Node_Term* tmp = (Node_Term*) old_child;
	new_terminals[ch] = new Node_Term(*tmp);
	new_terminals[ch]->set_id(get_new_id());
	Node_Term* this_child = (Node_Term*) new_terminals[ch];
	this_child->set_expert(expert);
	new_child->set_child(ch, this_child);
	this_child->set_parent(new_child);
  }
  return new_child;
}

void Tree::delete_new_nt(Node_NT* nd) {  
  Network* gate = nd->get_gate();
  delete gate;
  int num_children = nd->get_num_children();
  for(int ch = 0; ch < num_children; ch ++) {
	Node_Term* tm = (Node_Term*) nd->get_child(ch);
	Network* xprt = tm->get_expert();
	delete xprt;
	xprt = (Network*) NULL;
	delete tm;
	tm = (Node_Term*) NULL;
  }
  delete nd;
  nd = (Node_NT*) NULL;
}

void Tree::swap_nodes(Node* new_node, Node* old_node) {
  // replace old_node with new_node and 
  // update parent / child relationships
  Node_NT* pa = (Node_NT*) old_node->get_parent();
  int node_id = old_node->get_id();
  if(pa != NULL) {
	pa->replace_child(node_id, new_node);
	new_node->set_parent(pa);
  }
}  


Node_NT* Tree::split_this_node(real best_thresh, int best_feat,
								Node* old_nd, int n_points, 
								real** jpost, real** cpost, 
								real** in, real** targ) {
  
  Network* _gate;
  if(n_gates == 0)
	_gate = new_class_net(gate_ensemble, n_points, n_input, 2);
  else _gate = gates[0];

  Node_NT* new_node = (Node_NT*) create_new_nt(old_nd, 2, _gate);
 
  Node* term1 = new_node->get_child(0);
  Node* term2 = new_node->get_child(1);  
  
  new_node->reset_nodes();
  for(int n = 0; n < n_points; n ++) {
	term1->set_joint_post(n, jpost[0][n]);
	term2->set_joint_post(n, jpost[1][n]);  
	new_node->find_posts1(n);  
	new_node->reset_nodes();
  }
  //  term1->randomize_weights(1.0);
  //  term2->randomize_weights(1.0);  
  term1->init(n_points, in, targ);
  term2->init(n_points, in, targ);
  Network* gt = new_node->get_gate();
  //  gt->randomize_weights(1.0);  
  new_node->train_gates(n_points, in);
 
  return new_node;
}

real Tree::get_node_posts(Node* nod, int n_points, real** in, real** targ)
{
  real lik = 0.0;
  real ax = 0.0;
  if(strcmp(nod->get_class(), "Node_NT") == 0) {
	Node_NT* nd = (Node_NT*) nod;
	int n_child = nd->get_num_children();
	for(int n = 0; n < n_points; n ++) {
	  node->set_joint_prob(0.0, in[n], n);
	  real sum = LZERO;
	  real lik_n = LZERO;
	  for(int ch = 0; ch < n_child; ch ++) {
		Node_Term* child = (Node_Term*)(nd->get_child(ch));
		Network* expert = child->get_expert();
		real dens = expert->get_joint(in[n], targ[n]);
		real j = child->get_joint_prob(n);
		real jp = j + dens;
		lik_n = LAdd(lik_n, jp);
		child->set_joint_post(n, jp);
		if(jp > LSMALL)
		  sum = LAdd(sum, jp);
	  }
	  lik += lik_n;
	  for(int ch2 = 0; ch2 < n_child; ch2 ++) {	
		Node_Term* child = (Node_Term*)(nd->get_child(ch2));
		real jp = child->get_joint_post(n) - sum;
		jp += nd->get_joint_post(n);
		child->set_joint_post(n, jp);
		Network* expert = child->get_expert();
		real dens = expert->get_joint(in[n], targ[n]);
		real j = child->get_joint_prob(n);
		ax += Sexp(jp) * (j + dens);
	  }
	}
  }
  else {
	Node_Term* nd = (Node_Term*) nod;
	lik = 0.0;
	ax = 0.0;
	for(int n = 0; n < n_points; n ++) {
	  node->set_joint_prob(0.0, in[n], n);
	  Network* expert = nd->get_expert();
	  real dens = expert->get_joint(in[n], targ[n]);
	  real j = nd->get_joint_prob(n);
	  real j_dens = j + dens;
	  lik += j_dens;
	  real jp = nd->get_joint_post(n);
	  ax += Sexp(jp) * j_dens;
	}
  }
  return lik;
}

real Tree::get_node_aux(Node* nod, int n_points, real** in, real** targ)
{
  real lik = 0.0;
  real ax = 0.0;
  if(strcmp(nod->get_class(), "Node_NT") == 0) {
	Node_NT* nd = (Node_NT*) nod;
	int n_child = nd->get_num_children();
	for(int n = 0; n < n_points; n ++) {
	  node->set_joint_prob(0.0, in[n], n);
	  real sum = LZERO;
	  real lik_n = LZERO;
	  for(int ch = 0; ch < n_child; ch ++) {
		Node_Term* child = (Node_Term*)(nd->get_child(ch));
		Network* expert = child->get_expert();
		real dens = expert->get_joint(in[n], targ[n]);
		real j = child->get_joint_prob(n);
		real jp = j + dens;
		lik_n = LAdd(lik_n, jp);
		child->set_joint_post(n, jp);
		if(jp > LSMALL)
		  sum = LAdd(sum, jp);
	  }
	  lik += lik_n;
	  for(int ch2 = 0; ch2 < n_child; ch2 ++) {	
		Node_Term* child = (Node_Term*)(nd->get_child(ch2));
		real jp = child->get_joint_post(n) - sum;
		real pa_jp = nd->get_joint_post(n);
		jp += pa_jp;
		child->set_joint_post(n, jp);
		Network* expert = child->get_expert();
		real dens = expert->get_joint(in[n], targ[n]);
		real j = child->get_joint_prob(n);
		real j_dens = j + dens;
		ax += Sexp(jp) * (j_dens);
	  }
	}
  }
  else {
	Node_Term* nd = (Node_Term*) nod;
	lik = 0.0;
	ax = 0.0;
	for(int n = 0; n < n_points; n ++) {
	  node->set_joint_prob(0.0, in[n], n);
	  Network* expert = nd->get_expert();
	  real dens = expert->get_joint(in[n], targ[n]);
	  real j = nd->get_joint_prob(n);
	  real j_dens = j + dens;
	  lik += j_dens;
	  real jp = nd->get_joint_post(n);
	  ax += Sexp(jp) * j_dens;
	}
  }
  return ax;
}

real Tree::get_node_sum_jpost(Node* nod, int n_points, real** in, real** targ)
{
  real sum_jp = LZERO;
  if(strcmp(nod->get_class(), "Node_NT") == 0) {
	Node_NT* nd = (Node_NT*) nod;
	int n_child = nd->get_num_children();
	for(int n = 0; n < n_points; n ++) {
	  for(int ch2 = 0; ch2 < n_child; ch2 ++) {	
		Node_Term* child = (Node_Term*)(nd->get_child(ch2));
		real jp = child->get_joint_post(n);
		sum_jp = LAdd(sum_jp, jp);
	  }
	}
  }
  else {
	Node_Term* nd = (Node_Term*) nod;
	for(int n = 0; n < n_points; n ++) {
	  real jp = nd->get_joint_post(n);
	  sum_jp = LAdd(sum_jp, jp);
	}
  }
  return sum_jp;
}

void Tree::assign_probs_from_thresh_sig(Node* nd, int n_points,
									real** in, real best_thresh,
									int thresh_posn,
									int best_feat,
									real** jpost, real** cpost,
									int n_batch,
									real** srt_in,
									int* indx
									) {
  int start_posn = 0; // where the batch starts
  //  printf("size batch = %d\n", n_batch);
  //  printf("start_posn = %d %f\n", start_posn, srt_in[start_posn][best_feat]);
  //  printf("thresh_posn = %d\n", thresh_posn);
  int left_posn = (int)((floor) ((thresh_posn - start_posn) / 2.0));
  //  printf("left_posn = %d %f\n", left_posn, srt_in[left_posn][best_feat]);
  int mapped_left_posn = indx[left_posn + 1] - 1;
  real left_val = srt_in[left_posn][best_feat];
  //  if(thresh_posn - left_posn < 
  //  left_val = 
  real prob = 0.95;
  real logis = prob / (1.0 - prob);
  real denom = (left_val - best_thresh);
  if(fabs(denom) < 1e-7) denom = 1e-7;
  real w1 = Slog(logis) / denom;
  real w0 = - w1 * best_thresh;
  for(int n =0; n <n_points; n ++) {
	real a = w0 + w1 * in[n][best_feat];
	real p = 1.0 / ( 1.0 + Sexp(-a));
	real logp = Slog(p);
	real nlogp = Slog(1-p);
	real jp = nd->get_joint_post(n);
	jpost[0][n] = logp + jp;
	jpost[1][n] = nlogp + jp;
	cpost[n][0] = p;
	cpost[n][1] = 1 - p;
	/*	printf("%f %f %f %f %f %f %f \n", 
		   Sexp(jpost[0][n]), 
		   Sexp(jpost[1][n]),
		   cpost[n][0], 
		   cpost[n][1],
		   Sexp(jp),
		   best_thresh,
		   in[n][best_feat]);*/
  }
}

void Tree::assign_probs_from_thresh(Node* nd, int n_points,
							  real** in, real best_thresh,
							  int best_feat,
							  real** jpost, real** cpost
							  ) {
  for(int n =0; n <n_points; n ++) {
	real jp = nd->get_joint_post(n);
	if(in[n][best_feat] < best_thresh) {
	  jpost[0][n] = jp;
	  jpost[1][n] = LZERO + jp;
	  cpost[n][0] = 1.0;
	  cpost[n][1] = 0.0;
	}
	else {
	  jpost[1][n] = jp;
	  jpost[0][n] = LZERO + jp;
	  cpost[n][1] = 1.0;
	  cpost[n][0] = 0.0;
	  }
  }
}

real Tree::find_best_hard_thresh(Node* nd,
								 int batch_n, real** in_batch,
								 real** in, real** targ,
								 int* best_feat, real* max_diff,
								 int* best_thresh_posn) {
  *best_feat = -1;
  real best_thresh = 99.0;
  *max_diff = -FLT_MAX;
  real* init_pred = RealVector(n_output);
  real* left_pred = RealVector(n_output);
  real* right_pred = RealVector(n_output);
  for(int i = 1; i < n_input; i ++) {
	for(int n = 0; n < batch_n; n ++) {
	  tmp[n+1] = in_batch[n][i];
	}
	indexx(batch_n, tmp, indx);
	for(int n = 0; n < batch_n; n ++) {
	  srt_in[n][i] = tmp[indx[n+1]];
	}
	for(int o = 0; o < n_output; o ++) {
	  init_pred[o] = 0.0;
	}
	real init_pts = 0.01;
	for(int n = 0; n < batch_n; n ++) {
	  real wt = Sexp(nd->get_joint_prob(n));
	  for(int o = 0; o < n_output; o ++) {
		init_pred[o] += wt * targ[n][o];
	  }
	  init_pts += wt;
	}
	for(int o = 0; o < n_output; o ++) {
	  init_pred[o] /= init_pts;
	}
	real init_err = 0.0;
	for(int n = 0; n < batch_n; n ++) {
	  for(int o = 0; o < n_output; o ++) {
		real arg = init_pred[o] - targ[n][0];
		init_err += arg*arg;
	  }
	}
	init_err /= (double) batch_n;
	for(int n = 0; n < batch_n-1; n ++) {
	  real thresh = (srt_in[n][i] + srt_in[n+1][i]) / 2.0;
	  int thresh_posn = n+1;
	  for(int o = 0; o < n_output; o ++) {
		left_pred[o] = 0.0;
		right_pred[o] = 0.0;
	  }
	  real left_pts = 0.01;
	  real right_pts = 0.01;
	  for(int n2 = 0; n2 < n; n2 ++) {
		int n2a = indx[n2+1] -1;
		real wt = Sexp(nd->get_joint_prob(n2a));
		for(int o = 0; o < n_output; o ++) {
		  left_pred[o] += wt*targ[n2a][o];
		}
		left_pts += wt;
	  }
	  for(int o = 0; o < n_output; o ++) {
		left_pred[o] /= left_pts;
	  }
	  for(int n3 = n; n3 < batch_n; n3 ++) {
		int n5 = indx[n3+1] -1;
		real wt = Sexp(nd->get_joint_prob(n5));
		for(int o = 0; o < n_output; o ++) {
		  right_pred[o] += wt*targ[n5][o]; 
		}
		right_pts += wt;
	  }
	  for(int o = 0; o < n_output; o ++) {
		right_pred[o] /= right_pts;
	  }
	  real err = 0.0;
	  for(int n4 = 0; n4 < batch_n; n4 ++) {
		for(int o = 0; o < n_output; o ++) {
		  real pred = 0.0;
		  if(in[n4][i] > thresh) pred = right_pred[o];
		  else pred = left_pred[o];
		  real arg = (pred - targ[n][o]);
		  err += arg*arg;
		}
	  }
	  err /= (double) batch_n;
	  real diff = init_err - err;

	  if(diff > *max_diff) {
		*max_diff = diff;
		best_thresh = thresh;
		*best_thresh_posn = thresh_posn;
		*best_feat = i;
	  }
	  //	  printf("diff = %f, best = %f, feat = %d, best = %d, thresh = %f, best = %f\n", diff, *max_diff, i, *best_feat, thresh, best_thresh);
	}
  }
  FreeRealVector(&init_pred);
  FreeRealVector(&left_pred);  
  FreeRealVector(&right_pred);
  return best_thresh;
}

Node_NT* Tree::find_best_hard_node(Node* nd, int n_points,
								   real** in, real** targ, 
								   real* diff,
								   int* success) {

  Node_NT* new_nd = NULL;
  int batch_n = 0;
  for(int n = 0; n < n_points; n ++) {
	real wt = Sexp(nd->get_joint_post(n));
	if(wt > 0.01) {
	  for(int i = 0; i < n_input; i ++) {
		in_batch[batch_n][i] = in[n][i];
	  }
	  batch_n ++;
	}
  }
  if(batch_n < 4) {
	printf("no split tried, batch_n < 4\n");
	*success = 0;
  }
  else {
	int best_feat = -1;
	real best_thresh = 0.0;
	real max_diff = -FLT_MAX;
	int best_thresh_posn = -1;
	*success = 0;
	best_thresh = find_best_hard_thresh(nd,
										batch_n, 
										in_batch, 
										in, targ,
										&best_feat, 
										&max_diff,
										&best_thresh_posn);

	if(best_feat != -1) {
	  
	  real** srt_in = RealMatrix2(batch_n, n_input);
	  int* indx = IntVector(batch_n + 1 );
	  real* tmp = RealVector(batch_n + 1);
	  
	  for(int n = 0; n < batch_n; n ++) {
		tmp[n+1] = in_batch[n][best_feat];
	  }
	  indexx(batch_n, tmp, indx);
	  for(int n = 0; n < batch_n; n ++) {
		srt_in[n][best_feat] = tmp[indx[n+1]];
	  }
	  assign_probs_from_thresh_sig(nd, n_points, in, best_thresh, 
								   best_thresh_posn,
								   best_feat, jpost, cpost, batch_n,
								   srt_in, indx);
	  
	  
	  /*	assign_probs_from_thresh(nd, n_points,
			in, best_thresh, best_feat, 
			jpost, cpost);*/
	  
	  
	  real old_like = get_node_aux(nd, n_points, in, targ);
	  real sum_jp = Sexp(get_node_sum_jpost(nd, n_points, in, targ));
	  
	  new_nd =
	  split_this_node(best_thresh, best_feat, 
					  nd, n_points, jpost, cpost, in, targ);
	  real new_like = get_node_aux(new_nd, n_points, in, targ);
	  *diff = (new_like - old_like) / sum_jp;
	  //	  printf("%f %f diff = %f %d %f \n", new_like, old_like, *diff, new_nd->get_id(), best_thresh);
	  *success = 1;
	}
  }
  return new_nd;
}

void Tree::split_nodes_hard(int n_points, real** in, real** targ) {

  real max_diff = -FLT_MAX;
  int best_node_num = -1;
  Node** trial_node = new Node*[n_terminals]; // array of trial nodes
  int n_splits = 0;
  Node_NT* best_node;
  int best_split_num = -1;
  for(int term = 0; term < n_terminals; term ++) {
	Node* nd = term_nodes[term];
	printf("trying node %d %d\n", term, nd->get_id());
	real diff = -FLT_MAX;
	int success = 0;
	Node_NT* new_nd = find_best_hard_node(nd, n_points,
										  in,targ, 
										  &diff,
										  &success);
	if(diff > max_diff ) {
	  best_node_num = term;
	  max_diff = diff;
	  best_split_num = n_splits;
	  trial_node[n_splits] = new_nd;
	  best_node = new_nd;
	  n_splits ++;
	}
	else {
	  printf("diff = %f, max_diff = %f : deleting node\n", diff, max_diff);
	  delete new_nd;
	  new_nd = (Node_NT*) NULL;
	}
  }
  if(best_node_num != -1) {
	printf("best_node = %d id = %d\n", best_node_num, term_nodes[best_node_num]->get_id());
	
	Node_NT* new_nd = best_node;
	
	if(n_gates == 0) {
	  printf("setting new node to root\n");
	  node = new_nd;
	}
	else {
	  swap_nodes(new_nd, term_nodes[best_node_num]);
	}
  }
  for(int t = 0; t < n_splits; t ++) {
	if(t != best_split_num)
	  delete trial_node[t];
  }
  delete [] trial_node;
  delete [] term_nodes;
  delete [] terminals;
  delete [] gates;
  delete [] nt_nodes;
  gather_nodes();
  mac_pass = 0;
  macopt_free();
  int np = get_n_params();
  printf("nparams = %d\n", np);
}

void Tree::split_nodes_max_like(int n_points, real** in, real** targ) {


  // first find the best node....
  real max_like = -FLT_MAX;
  int ml_node = 0;
  for(int term = 0; term < n_terminals; term ++) {
	Node* nd = term_nodes[term];
	real like = get_node_aux(nd, n_points, in, targ);
	real sum_jp = get_node_sum_jpost(nd, n_points, in, targ);
	sum_jp = Sexp(sum_jp);
	like /= sum_jp;
	if(like > max_like) {
	  max_like = like;
	  ml_node = term;
	}
	printf("node %d, like %f\n", term, like);
  }
  Node* nd = term_nodes[ml_node];
  real diff; 
  int success = 0;
    Node_NT* best_node = find_best_soft_node(nd, n_points, in, targ,
										   &diff, 
										   &success);
  /*    Node_NT* best_node = find_best_hard_node(nd, n_points, in, targ,
										   &diff, 
										   &success);*/
  
  printf("best_node = %d id = %d\n", ml_node, term_nodes[ml_node]->get_id());
	
  Node_NT* new_nd = best_node;
  if(success == 1) {  
	if(n_gates == 0) {
	  printf("setting new node to root\n");
	  node = new_nd;
	}
	else {
	  swap_nodes(new_nd, term_nodes[ml_node]);
	}
  }
  else {
	printf("no success in splitting\n");
  }
  delete [] term_nodes;
  delete [] terminals;
  delete [] gates;
  delete [] nt_nodes;
  gather_nodes();
  mac_pass = 0;
  macopt_free();
  int np = get_n_params();
  printf("nparams = %d\n", np);
}

