#include "tree.h"

// -------------------------------------------------------
// training routines
// -------------------------------------------------------

void Tree::propagate_posts(Node** nodes, int num_nodes, int n) {
  for(int xp = 0 ; xp < num_nodes; xp ++) {
	Node_NT* pa = (Node_NT*) (nodes[xp]->get_parent());
	if(pa != NULL)
	  pa->find_posts2(n);
  }
  node->reset_nodes();
}

void Tree::get_posts(int n_points, real** in, real** targ)
{
  real over_sum = LZERO;
  real lik = 0.0;
  real jp_sum = LZERO;
  real ax = 0.0;
  for(int n = 0; n < n_points; n ++) {
	node->set_joint_prob(0.0, in[n], n);
	real sum = LZERO;
	real lik_n = LZERO;
	for(int ch = 0; ch < n_terminals; ch ++) {
	  real dens = terminals[ch]->get_joint(in[n], targ[n]);
	  real j = term_nodes[ch]->get_joint_prob(n);
	  jp_sum = LAdd(jp_sum, j);
	  real jp = j + dens;

	  lik_n = LAdd(lik_n, jp);
	  term_nodes[ch]->set_joint_post(n, jp);
	  if(jp > LSMALL)
		sum = LAdd(sum, jp);
	}
	lik += lik_n;
	for(int ch2 = 0; ch2 < n_terminals; ch2 ++) {	
	  real jp = term_nodes[ch2]->get_joint_post(n) - sum;
	  term_nodes[ch2]->set_joint_post(n, jp);
	  real dens = terminals[ch2]->get_joint(in[n], targ[n]);
	  real j = term_nodes[ch2]->get_joint_prob(n);
	  ax += Sexp(jp) * (j + dens);
	  over_sum = LAdd(over_sum, jp);
	}
	propagate_posts(term_nodes, n_terminals, n);
  }
  aux = ax;
  like = lik;
}

void Tree::get_posts2(int n_points, real** in, real** targ)
{
  real over_sum = LZERO;
  real lik = 0.0;
  real jp_sum = LZERO;
  real ax = 0.0;
  for(int n = 0; n < n_points; n ++) {
	node->reset_nodes();
	node->set_joint_prob(0.0, in[n], n);
	node->set_dens(n, in[n], targ[n]);
	node->set_posts(n);
	real lik_n = LZERO;
	for(int ch = 0; ch < n_terminals; ch ++) {
	  real dens = terminals[ch]->get_joint(in[n], targ[n]);
	  real j = term_nodes[ch]->get_joint_prob(n);
	  real jp = j + dens;
	  lik_n = LAdd(lik_n, jp);
	}
	lik += lik_n;
	for(int ch2 = 0; ch2 < n_terminals; ch2 ++) {	
	  real jp = term_nodes[ch2]->get_joint_post(n);
	  real dens = terminals[ch2]->get_joint(in[n], targ[n]);
	  real j = term_nodes[ch2]->get_joint_prob(n);
	  ax += Sexp(jp) * (j + dens);
	}
  }
  aux = ax;
  like = lik;
}


void Tree::get_hard_posts(int n_points, real** in, real** targ)
{
  for(int n = 0; n < n_points; n ++) {
	node->set_joint_prob(0.0, in[n], n);
	real sum = LZERO;
	for(int ch = 0; ch < n_terminals; ch ++) {
	  real dens = terminals[ch]->get_joint(in[n], targ[n]);
	  real jp = term_nodes[ch]->get_joint_prob(n) + dens;
	  term_nodes[ch]->set_joint_post(n, jp);
	  sum = LAdd(sum, jp);
	}
	real max_jp = -FLT_MAX;
	int best_term = -1;
	for(int ch2 = 0; ch2 < n_terminals; ch2 ++) {	
	  real jp = term_nodes[ch2]->get_joint_post(n) - sum;
	  if(jp > max_jp){
		max_jp = jp;
		best_term = ch2;
	  }
	  term_nodes[ch2]->set_joint_post(n, jp);
	}
	if(best_term == -1)
	  error(1, "best term is -1");
	for(int ch3 = 0; ch3 < n_terminals; ch3 ++) {	
	  term_nodes[ch3]->set_joint_post(n, LZERO);
	}
	term_nodes[best_term]->set_joint_post(n, 0.0);
	propagate_posts(term_nodes, n_terminals, n);
  }
}

void Tree::update_all(int n_points, real** in, real** targ)
{
  get_posts2(n_points, in, targ);

  if(debug > 3){
    for(int point2 = 0; point2 < n_points; point2 ++){
      node->print_posteriors(point2);
    }
  }
  node->train(n_points, in, targ);
}



int Tree::train(int n_epochs, 
				int init_epoch,
				real** inputs, 
				real** targets,
				FILE* fp_out, 
				int ntrain,
				int verb) {
  double log_like = 0.0;
  double aux_like = 0.0;
  real double_diff = 2.0;
  real old_diff = 0.0;
  real average_diff = 2.0;
  int average_length = 3;
  double diff = 0.0;
  real av = 0.0;
  int rec_epoch = init_epoch;
  int epoch = 0;
  time_t start_time;	// Time we started
  time_t stop_time;		// Time we stopped
  start_time = time(NULL);
  while(((fabs(average_diff) > 1e-5) && (epoch < n_epochs))) {
    if(epoch > 1) {
      if((epoch % average_length) == 0){
		average_diff = av / average_length;
		double_diff = fabs(old_diff - average_diff);
		old_diff = average_diff;
		av = 0.0;
      }
      av += diff;
    }
    double old_log_like = like;
	update_all(ntrain, inputs, targets);
	log_like = like;
	aux_like = aux;
	double error = compute_error(ntrain, inputs, targets);
    diff = fabs(log_like - old_log_like);
	if(diff < 0.01) {
	  get_posts2(ntrain, inputs, targets);
	  mac_train(ntrain, inputs, targets);
	}
	if(verb > 0)
	  printf("%4d\t%4.3g\t%4.3g\t%4.3g\t%4.3g\t%4.3g\t\n",
			 rec_epoch,
			 error,
			 log_like,
			 aux_like,
			 diff,
			 average_diff);
	
    epoch ++;
	rec_epoch ++;
  }
  stop_time = time(NULL);
  double kcps = (double) ntrain * (double) n_conn * (double) epoch/
	(double) (stop_time - start_time) * 0.001;
  printf("Speed = %4.3g Kcps\n", kcps);

  return rec_epoch;
}

int Tree::cv_train(int n_epochs, 
				   int init_epoch,
				   real** trn_inputs, 
				   real** trn_targets,
				   real** tst_inputs, 
				   real** tst_targets,
				   FILE* fp_out, 
				   int ntrain,
				   int ntest) {
  double diff = 0.1;
  int rec_epoch = init_epoch;
  int epoch = 0;
  int best_epoch = 1;
  double min_err = compute_error(ntest, tst_inputs, tst_targets);
  //  double min_err = FLT_MAX;
  double err = FLT_MAX;
  rewind(fp_out);
  write_weights(fp_out);
  while((epoch < 30) ||((epoch < best_epoch * 1.5) 
						&& (epoch < n_epochs))) {
    epoch ++;
	rec_epoch ++;
	if(err < min_err) {
	  // save current set of weights...
	  min_err = err;
	  rewind(fp_out);
	  write_weights(fp_out);
	  printf("(s)\n");
	  best_epoch = epoch;
	}
	else{
	  printf("\n");
	}
	// train on training set
	double old_like = like;
	get_posts2(ntrain, trn_inputs, trn_targets);
	double like_diff = like - old_like;
	node->train(ntrain, trn_inputs, trn_targets);
	if((like_diff < 0.01)) {
	  get_posts2(ntrain, trn_inputs, trn_targets);
	  mac_train(ntrain, trn_inputs, trn_targets);
	}
	double old_error = err;
	err = compute_error(ntest, tst_inputs, tst_targets);
	double trn_err = compute_error(ntrain, trn_inputs, trn_targets);
    diff = old_error - err;
    // output results

    printf("%4d\t%4.3g\t%4.3g\t%4.3g\t%d\t", 
		   rec_epoch,
		   err,
		   min_err,
		   trn_err,
		   best_epoch);
  }

  return rec_epoch;
}

int Tree::cv_train(int n_epochs, 
				   int init_epoch,
				   real** trn_inputs, 
				   real** trn_targets,
				   real** tst_inputs, 
				   real** tst_targets,
				   FILE* fp_out, 
				   int ntrain,
				   int ntest,
				   real* min_err) {
  double diff = 0.1;
  int rec_epoch = init_epoch;
  int epoch = 0;
  int best_epoch = 1;
  real err = compute_error(ntest, tst_inputs, tst_targets);
  while((epoch < 30) ||((epoch < best_epoch * 1.5) 
						&& (epoch < n_epochs))) {
    epoch ++;
	rec_epoch ++;
	if(err < *min_err) {
	  // save current set of weights...
	  *min_err = err;
	  rewind(fp_out);
	  write_weights(fp_out);
	  printf("(s)\n");
	  best_epoch = epoch;
	}
	else{
	  printf("\n");
	}
	// train on training set
	double old_like = like;
	get_posts2(ntrain, trn_inputs, trn_targets);
	double like_diff = like - old_like;
	node->train(ntrain, trn_inputs, trn_targets);
	if((like_diff < 0.01)) {
	  get_posts2(ntrain, trn_inputs, trn_targets);
	  mac_train(ntrain, trn_inputs, trn_targets);
	}
	double old_error = err;
	err = compute_error(ntest, tst_inputs, tst_targets);
	double trn_err = compute_error(ntrain, trn_inputs, trn_targets);
    diff = old_error - err;
    // output results

    printf("%4d\t%4.3g\t%4.3g\t%4.3g\t%d\t", 
		   rec_epoch,
		   err,
		   *min_err,
		   trn_err,
		   best_epoch);
  }

  return rec_epoch;
}
